package com.hedera.mirror.grpc.repository;

/*-
 * ‌
 * Hedera Mirror Node
 * ​
 * Copyright (C) 2019 Hedera Hashgraph, LLC
 * ​
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ‍
 */

import javax.inject.Named;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.data.r2dbc.core.DatabaseClient;
import org.springframework.data.r2dbc.query.Criteria;
import reactor.core.publisher.Flux;

import com.hedera.mirror.grpc.converter.InstantToLongConverter;
import com.hedera.mirror.grpc.domain.TopicMessage;
import com.hedera.mirror.grpc.domain.TopicMessageFilter;

@Log4j2
@Named
@RequiredArgsConstructor
public class TopicMessageRepositoryCustomImpl implements TopicMessageRepositoryCustom {

    private final DatabaseClient databaseClient;
    private final InstantToLongConverter instantToLongConverter;

    @Override
    public Flux<TopicMessage> findByFilter(TopicMessageFilter filter) {
        Criteria whereClause = Criteria.where("realm_num")
                .is(filter.getRealmNum())
                .and("topic_num")
                .is(filter.getTopicNum())
                .and("consensus_timestamp")
                .greaterThanOrEquals(instantToLongConverter.convert(filter.getStartTime()));

        if (filter.getEndTime() != null) {
            whereClause = whereClause.and("consensus_timestamp")
                    .lessThan(instantToLongConverter.convert(filter.getEndTime()));
        }

        Pageable pageable = filter.hasLimit() ? PageRequest.of(0, (int) filter.getLimit()) : Pageable.unpaged();

        return databaseClient.select()
                .from(TopicMessage.class)
                .matching(whereClause)
                .orderBy(Sort.by("consensus_timestamp"))
                .page(pageable)
                .fetch()
                .all()
                .name("findByFilter")
                .metrics()
                .doOnSubscribe(s -> log.debug("Executing query: {}", filter))
                .doOnCancel(() -> log.debug("[{}] Cancelled query", filter.getSubscriberId()))
                .doOnComplete(() -> log.debug("[{}] Completed query", filter.getSubscriberId()))
                .doOnNext(t -> log.trace("[{}] Next message: {}", filter.getSubscriberId(), t));
    }
}
